@@ -1,6 +1,7 @@
 mod contextual;
 mod funcs;
 mod generics;
+mod is_uninhabited;
 mod named;
 mod ns_path;
 mod resolve_target_type;
@@ -9,14 +10,15 @@ mod traits;
 mod type_ref_like;
 
 use crate::codegen::funcs::{
-    fn_types, render_raw_return_to_js, AccessType, Constructor, FnPrototypeExt, HasFnPrototype,
-    InternalFunc, PropertyAccessor, WrapperFunc,
+    fn_types, render_exposed_to_js_wrapper_closure, render_raw_return_to_js, AccessType,
+    Constructor, FnPrototypeExt, HasFnPrototype, InternalFunc, PropertyAccessor, WrapperFunc,
 };
 use crate::codegen::generics::{
     apply_type_params, render_type_params, render_type_params_with_constraints,
     render_type_params_with_lifetimes, ResolveGeneric, TypeEnvImplying,
 };
-use crate::codegen::named::{type_name, CasedTypeIdent, FnOverloadName, Named, SimpleNamed};
+use crate::codegen::is_uninhabited::IsUninhabited;
+use crate::codegen::named::{CasedTypeIdent, FnOverloadName, Named, SimpleNamed, UnionCaseName};
 use crate::codegen::ns_path::ToNsPath;
 use crate::codegen::resolve_target_type::ResolveTargetType;
 use crate::codegen::serialization_type::{SerializationType, SerializationTypeGetter};
@@ -211,37 +213,6 @@ fn get_recursive_fields_with_type_params(
     our_fields.chain(super_fields).collect()
 }
 
-trait IsUninhabited {
-    fn is_uninhabited(&self) -> bool;
-}
-
-impl IsUninhabited for TargetEnrichedTypeInfo {
-    fn is_uninhabited(&self) -> bool {
-        match self {
-            TargetEnrichedTypeInfo::Ref(r) => r.is_uninhabited(),
-            TargetEnrichedTypeInfo::Union(Union { types, .. }) => {
-                types.iter().all(IsUninhabited::is_uninhabited)
-            }
-            _ => false,
-        }
-    }
-}
-
-impl IsUninhabited for TypeRef {
-    fn is_uninhabited(&self) -> bool {
-        matches!(
-            self.referent,
-            TypeIdent::Builtin(Builtin::PrimitiveNull)
-                | TypeIdent::Builtin(Builtin::PrimitiveUndefined)
-                | TypeIdent::Builtin(Builtin::PrimitiveVoid),
-        )
-    }
-}
-
-fn type_to_union_case_name(typ: &TargetEnrichedTypeInfo) -> Identifier {
-    type_name(typ)
-}
-
 fn path_relative_to_cargo_toml<T: AsRef<Path>>(path: T) -> PathBuf {
     let mut best: Option<PathBuf> = None;
     let mut current_path: Option<PathBuf> = None;
@@ -273,6 +244,14 @@ trait FieldCountGetter {
     fn get_field_count(&self) -> usize;
 }
 
+impl FieldCountGetter for TypeRef {
+    fn get_field_count(&self) -> usize {
+        self.resolve_target_type()
+            .map(|t| t.get_field_count())
+            .unwrap_or(0usize)
+    }
+}
+
 impl FieldCountGetter for TargetEnrichedType {
     fn get_field_count(&self) -> usize {
         self.info.get_field_count()
@@ -326,9 +305,7 @@ impl FieldCountGetter for TargetEnrichedTypeInfo {
 }
 
 trait MemberContainer {
-    fn undefined_and_standard_members(
-        &self,
-    ) -> (Vec<&TargetEnrichedTypeInfo>, Vec<&TargetEnrichedTypeInfo>);
+    fn undefined_and_standard_members(&self) -> (Vec<&TypeRef>, Vec<&TypeRef>);
 
     fn has_undefined_member(&self) -> bool {
         !self.undefined_and_standard_members().0.is_empty()
@@ -336,17 +313,17 @@ trait MemberContainer {
 }
 
 impl MemberContainer for Union {
-    fn undefined_and_standard_members(
-        &self,
-    ) -> (Vec<&TargetEnrichedTypeInfo>, Vec<&TargetEnrichedTypeInfo>) {
-        self.types.iter().partition(|t| match t {
-            TargetEnrichedTypeInfo::Ref(t)
-                if t.referent == TypeIdent::Builtin(Builtin::PrimitiveUndefined) =>
-            {
-                true
-            }
-            _ => false,
-        })
+    fn undefined_and_standard_members(&self) -> (Vec<&TypeRef>, Vec<&TypeRef>) {
+        self.types
+            .iter()
+            .partition(|t| match t.resolve_target_type() {
+                Some(TargetEnrichedTypeInfo::Ref(t))
+                    if t.referent == TypeIdent::Builtin(Builtin::PrimitiveUndefined) =>
+                {
+                    true
+                }
+                _ => false,
+            })
     }
 }
 
@@ -532,7 +509,7 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                 let member_cases = not_undefined_members
                     .iter()
                     .map(|t| {
-                        let case = type_to_union_case_name(t);
+                        let case = t.union_case_name();
                         let attrs = match t.serialization_type() {
                             SerializationType::JsValue => {
                                 quote! {
@@ -566,7 +543,7 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                             )
                         } else {
                             let t = OwnedTypeInfo {
-                                type_info: t,
+                                type_info: &TargetEnrichedTypeInfo::Ref((*t).clone()),
                             };
                             (
                                 case.clone(),
@@ -578,7 +555,7 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                         }
                     })
                     .chain(undefined_members.iter().map(|t| {
-                        let case = type_to_union_case_name(t);
+                        let case = t.union_case_name();
 
                         (
                             case.clone(),
@@ -602,9 +579,10 @@ impl<'a, FS: Fs + ?Sized> ToTokens for WithFs<'a, TargetEnrichedType, FS> {
                 let impl_fns: Vec<_> = not_undefined_members
                     .iter()
                     .filter_map(|t| {
-                        let case = type_to_union_case_name(t).to_snake_case();
-                        let serialize_fn = render_serialize_fn(&case, t);
-                        let deserialize_fn = render_deserialize_fn(&case, t);
+                        let case = t.union_case_name().to_snake_case();
+                        let typ = t.resolve_target_type()?;
+                        let serialize_fn = render_serialize_fn(&case, &typ);
+                        let deserialize_fn = render_deserialize_fn(&case, &typ);
                         serialize_fn.zip(deserialize_fn).map(|(s, d)| {
                             quote! {
                                 #s
@@ -1592,8 +1570,8 @@ fn render_deserialize_fn(
         let name = p.rust_name();
         if p.is_variadic() {
             quote! {
-                for #name in #name.values().into_iter(){
-                    #args.push(&#name?);
+                for #name in #name.into_iter(){
+                    #args.push(&#name);
                 }
             }
         } else {
@@ -1654,8 +1632,8 @@ fn render_serialize_fn(
         }
     };
     let serialize_fn_name = serialize_field_name(field_name);
-    let invocation = f.invoke_with_name(field_name);
-    let closure = f.exposed_to_js_wrapped_closure(invocation, None);
+    let closure = render_exposed_to_js_wrapper_closure(f, field_name, None);
+    let closure_name = field_name.suffix_name("_closure");
     Some(quote! {
         #[allow(non_snake_case)]
         fn #serialize_fn_name<S>(#field_name: &#rendered_type, serializer: S) -> std::result::Result<S::Ok, S::Error>
@@ -1663,10 +1641,223 @@ fn render_serialize_fn(
             S: serde::ser::Serializer,
         {
             let #field_name = #field_name.clone();
-            let #field_name = #closure;
-            let jsv = ts_bindgen_rt::serialize_as_jsvalue(serializer, &#field_name.into_js_value());
-            //#field_name.forget(); // TODO: how do we properly handle memory management?
+            let #closure_name = #closure;
+            let jsv = ts_bindgen_rt::serialize_as_jsvalue(serializer, &#closure_name.into_js_value());
+            //#closure_name.forget(); // TODO: how do we properly handle memory management?
             jsv
         }
     })
 }
+
+#[cfg(test)]
+mod test {
+    use crate::{generate_rust_for_typescript, Error, MemFs};
+    use std::path::Path;
+
+    fn ts_to_rust(code: &str) -> Result<String, Error> {
+        let fs = {
+            let mut fs: MemFs = Default::default();
+            fs.set_cwd(Path::new("/"));
+            fs.add_file_at(Path::new("/test.d.ts"), code.to_string());
+            fs
+        };
+
+        Ok(generate_rust_for_typescript(fs, "/test")?.to_string())
+    }
+
+    #[test]
+    fn test_fn_tuple_spread_args() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export declare function foo(...args: [string, number]): void;
+        "#,
+        )?;
+
+        assert!(rust.replace(" ", "").contains("vec![]"));
+        assert!(rust.replace(" ", "").contains("JsValue::from"));
+        assert!(rust.replace(" ", "").contains("into_boxed_slice"));
+        Ok(())
+    }
+
+    #[test]
+    fn test_fn_union_tuple_spread_args() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export declare function foo(...args: [string, number] | [number, string]): void;
+        "#,
+        )?;
+
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"match args".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"FooParamsArgsUnion0(args) => {".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"FooParamsArgsUnion1(args) => {".replace(" ", "")));
+        assert!(rust.replace(" ", "").contains("into_boxed_slice()"));
+        Ok(())
+    }
+
+    #[test]
+    fn test_closure_tuple_spread_args() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export type foo = (...args: [string, number] | [number, string]) => void;
+        "#,
+        )?;
+
+        assert!(rust.replace(" ", "").contains(
+            &"pub type foo = dyn Fn(FooAliasedParamsArgs) -> std::result::Result<(), JsValue>;"
+                .replace(" ", "")
+        ));
+        Ok(())
+    }
+
+    #[test]
+    fn test_fn_spread_args() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export declare function foo(...args: Array<string>): void;
+        "#,
+        )?;
+
+        // aiming to ensure that we are creating an array to pass variadic args
+        assert!(rust.replace(" ", "").contains("collect::<Vec<_>>"));
+        assert!(rust.replace(" ", "").contains("into_boxed_slice"));
+        Ok(())
+    }
+
+    #[test]
+    fn test_closure_spread_args() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export type foo = (...args: Array<string>) => void;
+        "#,
+        )?;
+
+        assert!(rust.replace(" ", "").contains(
+            &"dyn Fn (Vec<String>) -> std::result::Result<(), JsValue>".replace(" ", "")
+        ));
+        Ok(())
+    }
+
+    #[test]
+    fn test_interface_function() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export interface Foo {
+                onSomething: Function;
+            }
+        "#,
+        )?;
+
+        assert!(rust.replace(" ", "").contains(
+            &"fn on_something(&self) -> std::result::Result<std::rc::Rc<dyn Fn(Vec<JsValue>) -> std::result::Result<JsValue, JsValue>>, JsValue>".replace(" ", "")
+        ));
+        assert!(rust.replace(" ", "").contains(
+            &"pub on_something: std::rc::Rc<dyn Fn(Vec<JsValue>) -> std::result::Result<JsValue, JsValue>>".replace(" ", "")
+        ));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"let _Args = js_sys::Array::new();".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"f.apply(&JsValue::null(), &_Args)".replace(" ", "")));
+        Ok(())
+    }
+
+    #[test]
+    fn test_interface_function_with_return() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export interface FooRet {
+                n: number;
+            }
+            export interface Foo {
+                onSomething(...args: Array<number>): FooRet;
+            }
+        "#,
+        )?;
+
+        assert!(rust.replace(" ", "").contains(
+            &"fn on_something(&self) -> std::result::Result<std::rc::Rc<dyn Fn(Vec<f64>) -> std::result::Result<FooRet, JsValue>>, JsValue>".replace(" ", "")
+        ));
+        assert!(rust.replace(" ", "").contains(
+            &"pub on_something: std::rc::Rc<dyn Fn(Vec<f64>) -> std::result::Result<FooRet, JsValue>>".replace(" ", "")
+        ));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"let _Args = js_sys::Array::new()".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"f.apply(&JsValue::null(), &_Args)".replace(" ", "")));
+        Ok(())
+    }
+
+    #[test]
+    fn test_class_variadic_method() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export declare class Foo {
+                onSomething(...s: (string | number)[]): void;
+            }
+        "#,
+        )?;
+
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl std::convert::From<Foo> for JsValue".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl std::convert::AsRef<JsValue> for Foo".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl wasm_bindgen::JsCast for Foo".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl wasm_bindgen::describe::WasmDescribe for Foo".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl wasm_bindgen::convert::IntoWasmAbi for Foo".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl wasm_bindgen::convert::FromWasmAbi for Foo".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl<'a> wasm_bindgen::convert::IntoWasmAbi for &'a Foo".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl serde::ser::Serialize for Foo".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"impl<'de> serde::de::Deserialize<'de> for Foo".replace(" ", "")));
+        assert!(rust.replace(" ", "").contains(
+            &"self.0.__TSB_on_something_FnVecOfFooOnSomethingParamsSTo(s.into_iter().map(|s_item| ts_bindgen_rt::to_jsvalue(&s_item).map_err(ts_bindgen_rt::Error::from).map_err(JsValue::from)).collect::<std::result::Result<Vec<_>, _>>().map_err(ts_bindgen_rt::Error::from).map_err(JsValue::from)?.into_boxed_slice())".replace(" ", "")
+        ));
+        Ok(())
+    }
+
+    #[test]
+    fn test_class_callback_method() -> Result<(), Error> {
+        let rust = ts_to_rust(
+            r#"
+            export declare class Foo {
+                onSomething(f: Function): void;
+            }
+        "#,
+        )?;
+
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"let arg0 = js_sys::Array::new()".replace(" ", "")));
+        assert!(rust
+            .replace(" ", "")
+            .contains(&"arg0.push(&_Variadic0_arg0);".replace(" ", "")));
+        assert!(rust.replace(" ", "").contains(
+            &"let result = f({ let mut arg0_vec = vec![]; for arg0_item in arg0.iter() { arg0_vec.push(arg0_item); } arg0_vec })?".replace(" ", "")
+        ));
+        Ok(())
+    }
+}
