@@ -1,12 +1,13 @@
 use crate::codegen::funcs::{AccessType, FnPrototypeExt, PropertyAccessor};
 use crate::codegen::generics::{render_type_params, render_type_params_with_constraints};
 use crate::codegen::generics::{ResolveGeneric, TypeEnvImplying};
-use crate::codegen::named::Named;
+use crate::codegen::named::{FnOverloadName, Named};
 use crate::codegen::resolve_target_type::ResolveTargetType;
+use crate::codegen::serialization_type::clone_item_of_type;
 use crate::identifier::{to_snake_case_ident, Identifier};
 use crate::ir::{
-    Class, Context, Func, Interface, Intersection, Member, TargetEnrichedTypeInfo, TypeIdent,
-    TypeParamConfig, TypeRef,
+    Builtin, Class, Context, Func, Interface, Intersection, Member, TargetEnrichedTypeInfo,
+    TypeIdent, TypeParamConfig, TypeRef,
 };
 use proc_macro2::TokenStream as TokenStream2;
 use quote::{format_ident, quote};
@@ -95,7 +96,21 @@ where
         let name = to_snake_case_ident(&n);
         match m {
             Member::Constructor(_) => Default::default(),
-            Member::Method(f) => vec![TraitMember::Method { name, method: f }],
+            Member::Method(f) => {
+                let is_overloaded = f.overloads.len() > 1;
+                f.overloads
+                    .iter()
+                    // TODO: name needs to be modified based on method
+                    .map(|o| TraitMember::Method {
+                        name: if is_overloaded {
+                            o.overload_name(&name)
+                        } else {
+                            name.clone()
+                        },
+                        method: o.clone(),
+                    })
+                    .collect()
+            }
             Member::Property(t) => {
                 let getter = PropertyAccessor {
                     property_name: name.clone(),
@@ -198,7 +213,7 @@ where
     };
 
     quote! {
-        #[allow(non_camel_case_types)]
+        #[allow(non_camel_case_types, non_snake_case)]
         #vis trait #trait_name #tps #super_decl {
             #(#method_decls)*
         }
@@ -332,8 +347,17 @@ impl Traitable for Interface {
             }
             TraitMember::Getter { prop, .. } => {
                 let property_name = &prop.property_name;
+                let typ = prop.typ.resolve_target_type().unwrap_or_else(|| {
+                    TargetEnrichedTypeInfo::Ref(TypeRef {
+                        referent: TypeIdent::Builtin(Builtin::PrimitiveAny),
+                        type_params: Default::default(),
+                        context: prop.typ.context.clone(),
+                    })
+                });
+                let cloned = clone_item_of_type(quote! { self.#property_name }, &typ);
+
                 quote! {
-                    Ok(self.#property_name.clone())
+                    Ok(#cloned)
                 }
             }
             TraitMember::Setter { prop, .. } => {
@@ -383,8 +407,6 @@ impl Traitable for Class {
         let name = trait_member.name();
         let name = &name.in_namespace(&cn);
         let slf = quote! { self };
-        let target = quote! { &target };
-        let mut_target = quote! { &mut target };
         let is_direct_member = member_defn_source
             .resolve_target_type()
             .map(|t| match t {
@@ -392,11 +414,23 @@ impl Traitable for Class {
                 _ => false,
             })
             .unwrap_or(false);
+        let target = if is_direct_member {
+            quote! { target }
+        } else {
+            quote! { &target }
+        };
+        // TODO: maybe we should require mut on setters in classes?
+        // if so, we can use the mut_target and tgt_mut should be as below
+        /*
+        let mut_target = quote! { &mut target };
         let tgt_mut = if matches!(trait_member, TraitMember::Setter { .. }) {
             quote! { mut }
         } else {
             quote! {}
         };
+        */
+        let mut_target = target.clone();
+        let tgt_mut = quote! {};
         let conv = if is_direct_member {
             quote! {
                 let #tgt_mut target = #slf;
