@@ -409,7 +409,7 @@ macro_rules! combine_effects {
 
 mod effect_mappers {
     use super::Effect;
-    use heck::TitleCase;
+    use heck::CamelCase;
 
     pub fn prepend_name<T: AsRef<str>>(prefix: T) -> impl Fn(Effect) -> Effect {
         move |e: Effect| match e {
@@ -420,7 +420,11 @@ mod effect_mappers {
                 typ,
                 generated_name_id,
             } => Effect::CreateType {
-                name: format!("{}_{}", prefix.as_ref().to_title_case(), &name),
+                name: if name.is_empty() {
+                    prefix.as_ref().to_camel_case()
+                } else {
+                    format!("{}_{}", prefix.as_ref().to_camel_case(), &name)
+                },
                 file,
                 ns,
                 typ,
@@ -918,7 +922,7 @@ impl From<Namespaced<Box<TypeInfoIR>>> for EffectContainer<Box<FlattenedTypeInfo
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Union {
-    pub types: Vec<FlattenedTypeInfo>,
+    pub types: Vec<TypeRef>,
 }
 
 impl ApplyNames for Union {
@@ -940,10 +944,17 @@ impl From<Namespaced<UnionIR>> for EffectContainer<Union> {
                 .types
                 .into_iter()
                 .map(|t| ns.in_ns(t))
-                .map(EffectContainer::from)
+                .enumerate()
+                .map(|(i, t)| {
+                    let ec = EffectContainer::from(t);
+                    combine_effects!(
+                        ec => (effect_mappers::prepend_name(i.to_string()));
+                        ec
+                    )
+                })
                 .collect();
             combine_effects!(
-                types => (effect_mappers::identity());
+                types => (effect_mappers::prepend_name("Union"));
                 Union {
                     types,
                 }
@@ -954,7 +965,7 @@ impl From<Namespaced<UnionIR>> for EffectContainer<Union> {
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Intersection {
-    pub types: Vec<FlattenedTypeInfo>,
+    pub types: Vec<TypeRef>,
 }
 
 impl ApplyNames for Intersection {
@@ -976,10 +987,17 @@ impl From<Namespaced<IntersectionIR>> for EffectContainer<Intersection> {
                 .types
                 .into_iter()
                 .map(|t| ns.in_ns(t))
-                .map(EffectContainer::from)
+                .enumerate()
+                .map(|(i, t)| {
+                    let ec = EffectContainer::from(t);
+                    combine_effects!(
+                        ec => (effect_mappers::prepend_name(i.to_string()));
+                        ec
+                    )
+                })
                 .collect();
             combine_effects!(
-                types => (effect_mappers::identity());
+                types => (effect_mappers::prepend_name("Intersection"));
                 Intersection {
                     types,
                 }
@@ -990,7 +1008,7 @@ impl From<Namespaced<IntersectionIR>> for EffectContainer<Intersection> {
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Tuple {
-    pub types: Vec<FlattenedTypeInfo>,
+    pub types: Vec<TypeRef>,
 }
 
 impl ApplyNames for Tuple {
@@ -1012,10 +1030,17 @@ impl From<Namespaced<TupleIR>> for EffectContainer<Tuple> {
                 .types
                 .into_iter()
                 .map(|t| ns.in_ns(t))
-                .map(EffectContainer::from)
+                .enumerate()
+                .map(|(i, t)| {
+                    let ec = EffectContainer::from(t);
+                    combine_effects!(
+                        ec => (effect_mappers::prepend_name(i.to_string()));
+                        ec
+                    )
+                })
                 .collect();
             combine_effects!(
-                types => (effect_mappers::identity());
+                types => (effect_mappers::prepend_name("Tuple"));
                 Tuple {
                     types,
                 }
@@ -1026,7 +1051,7 @@ impl From<Namespaced<TupleIR>> for EffectContainer<Tuple> {
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct TypeParamConfig {
-    pub constraint: Option<FlattenedTypeInfo>,
+    pub constraint: Option<TypeRef>,
     pub default_type_arg: Option<TypeRef>,
 }
 
@@ -1089,9 +1114,15 @@ fn combine_to_union(a: &TypeInfoIR, b: &TypeInfoIR) -> TypeInfoIR {
         (t1, TypeInfoIR::Union(u2)) => TypeInfoIR::Union(UnionIR {
             types: u2.types.iter().chain(iter::once(t1)).cloned().collect(),
         }),
-        (t1, t2) => TypeInfoIR::Union(UnionIR {
-            types: vec![t1.clone(), t2.clone()],
-        }),
+        (t1, t2) => {
+            if t1 == t2 {
+                t1.clone()
+            } else {
+                TypeInfoIR::Union(UnionIR {
+                    types: vec![t1.clone(), t2.clone()],
+                })
+            }
+        }
     }
 }
 
@@ -1241,7 +1272,7 @@ impl From<Namespaced<ParamIR>> for EffectContainer<Param> {
             let type_info = ns.in_ns(v.type_info).into();
 
             combine_effects!(
-                type_info => (effect_mappers::prepend_name(v.name.clone() + "Param"));
+                type_info => (effect_mappers::prepend_name(v.name.clone()));
                 Param {
                     name: v.name,
                     type_info,
@@ -1654,7 +1685,7 @@ pub fn flatten_types<Ts: IntoIterator<Item = TypeIR>>(types: Ts) -> impl Iterato
 
 #[cfg(test)]
 mod test {
-    use super::{flatten_types, FlatType, TypeIdent};
+    use super::{flatten_types, Builtin, FlatType, FlattenedTypeInfo, TypeIdent};
     use crate::fs::MemFs;
     use crate::parse::TsTypes;
     use crate::ArcFs;
@@ -1703,4 +1734,127 @@ mod test {
         // into real names. this avoids having to write some generic mapper.
         assert!(!debug.contains("GeneratedName"));
     }
+
+    #[test]
+    fn test_flatten_union_members() {
+        let code = r#"
+            export declare function myFunc(a: { n: string } | number);
+        "#;
+
+        let types = get_types_for_code(code).unwrap();
+
+        let i = types.get(&TypeIdent::Name {
+            name: "MyFunc_Params_A_Union_0".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            i.map(|i| &i.info),
+            Some(FlattenedTypeInfo::Interface(_)),
+        ));
+
+        let u = types.get(&TypeIdent::Name {
+            name: "MyFunc_Params_A".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            u.map(|u| &u.info),
+            Some(FlattenedTypeInfo::Union(_)),
+        ));
+    }
+
+    #[test]
+    fn test_flatten_intersection_members() {
+        let code = r#"
+            export declare function myFunc(a: { s: string } & { n: number });
+        "#;
+
+        let types = get_types_for_code(code).unwrap();
+
+        let i = types.get(&TypeIdent::Name {
+            name: "MyFunc_Params_A_Intersection_0".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            i.map(|i| &i.info),
+            Some(FlattenedTypeInfo::Interface(_)),
+        ));
+
+        let i = types.get(&TypeIdent::Name {
+            name: "MyFunc_Params_A_Intersection_1".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            i.map(|i| &i.info),
+            Some(FlattenedTypeInfo::Interface(_)),
+        ));
+
+        let i = types.get(&TypeIdent::Name {
+            name: "MyFunc_Params_A".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            i.map(|i| &i.info),
+            Some(FlattenedTypeInfo::Intersection(_)),
+        ));
+    }
+
+    #[test]
+    fn test_flatten_tuple_members() {
+        let code = r#"
+            export type A = [{ n: number }];
+        "#;
+
+        let types = get_types_for_code(code).unwrap();
+
+        let t = types.get(&TypeIdent::Name {
+            name: "A_Aliased".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            t.map(|t| &t.info),
+            Some(FlattenedTypeInfo::Tuple(_)),
+        ));
+
+        let i = types.get(&TypeIdent::Name {
+            name: "A_Aliased_Tuple_0".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            i.map(|i| &i.info),
+            Some(FlattenedTypeInfo::Interface(_)),
+        ));
+
+        let a = types.get(&TypeIdent::Name {
+            name: "A".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+        assert!(matches!(
+            a.map(|a| &a.info),
+            Some(FlattenedTypeInfo::Alias(_)),
+        ));
+    }
+
+    #[test]
+    fn test_widened_fn() {
+        let code = r#"
+            export declare function a(n: number): void;
+            export declare function a(s: string): void;
+        "#;
+
+        let types = get_types_for_code(code).unwrap();
+
+        let t = types.get(&TypeIdent::Name {
+            name: "a".to_string(),
+            file: Path::new("/test.d.ts").to_path_buf(),
+        });
+
+        if let FlattenedTypeInfo::FuncGroup(fg) = &t.unwrap().info {
+            assert_eq!(
+                fg.widened_fn.return_type.referent,
+                TypeIdent::Builtin(Builtin::PrimitiveVoid)
+            );
+        } else {
+            assert!(false);
+        }
+    }
 }
