@@ -327,7 +327,7 @@ from_struct!(
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct TypeParamConfig {
-    pub constraint: Option<TargetEnrichedTypeInfo>,
+    pub constraint: Option<TypeRef>,
     pub default_type_arg: Option<TypeRef>,
     pub context: Context,
 }
@@ -376,7 +376,7 @@ from_struct!(
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Union {
-    pub types: Vec<TargetEnrichedTypeInfo>,
+    pub types: Vec<TypeRef>,
     pub context: Context,
 }
 
@@ -387,7 +387,7 @@ from_struct!(
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Tuple {
-    pub types: Vec<TargetEnrichedTypeInfo>,
+    pub types: Vec<TypeRef>,
     pub context: Context,
 }
 
@@ -398,7 +398,7 @@ from_struct!(
 
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct Intersection {
-    pub types: Vec<TargetEnrichedTypeInfo>,
+    pub types: Vec<TypeRef>,
     pub context: Context,
 }
 
